
import curses, string, os, re
import asyncore
from TextEntry import TextEntry
from TextScroll import TextScroll

class StatusBar:
    "very simple single-line status bar that sits between the scrolling text region and the entry box"
    def __init__(self, win, y=23, maxx=78, text=''):
        self.win = win
        self.text = text
        self.relocate(y, maxx)

    def redraw(self):
        try:
            self.win.addstr(self.y, 0, ' ' * self.maxx, curses.A_REVERSE)
            list = string.split(self.text, '\t')
            self.win.addstr(self.y, 0, list[0][:self.maxx], curses.A_REVERSE)
            if len(list) > 1:
                self.win.addstr(self.y, self.maxx-len(list[1]), list[1][:self.maxx], curses.A_REVERSE)
        except curses.error:
            pass

    def relocate(self, y, maxx):
        self.y = y
        self.maxx = maxx
        self.redraw()

    def set_text(self, text):
        self.text = text
        self.redraw()


class FakeSocket:
    def __init__(self, fd=0):
        self.fd = fd
    def fileno(self):
        return self.fd
    def close(self):
        os.close(self.fd)


def make_printable(str):
    _r = '\006V+\006'
    _u = '\006V-\006'
    str = re.sub(r'([\x00-\x1F])', lambda x,_r=_r,_u=_u: _r + chr(ord(x.group(1))+64) + _u, str)
    str = re.sub(r'([\x7F])', lambda x,_r=_r,_u=_u: _r + '?' + _u, str)
    str = re.sub(r'([\x9F])', lambda x,_r=_r,_u=_u: _r + '/' + _u, str)
    str = re.sub(r'([\xA0])', lambda x,_r=_r,_u=_u: _r + ' ' + _u, str)
    str = re.sub(r'([\x80-\x9E])', lambda x,_r=_r,_u=_u: _r + chr(ord(x.group(1))-32) + _u, str)
    return str

class Console (asyncore.dispatcher):
    """
    TextEntry and TextScroll combined, with a StatusBar between them
    it's also a "dispatcher" for stdin, so that the asyncore poll loop will trigger it.
    """

    def __init__(self, stdscr, minheight=2, maxheight=5):
        asyncore.dispatcher.__init__(self)
        self.set_socket(FakeSocket(0))
        self.win = stdscr
        self.active = 1
        self.quote_mode = 0
        self.insert_mode = 1
        self.status_line = '%(more)s say2.py\t%(quo)s %(ovr)s :) '
        self.status_table = { }

        self.entry = TextEntry(self.win, 1)
        self.entry.set_expand(minheight, maxheight)

        # extra keystrokes
        self.entry.keymap['meta1-\x18'] = self.action_reset

        # set up signal handlers
        self.entry.signals['quote-on'] = self.signal_quote_on
        self.entry.signals['quote-off'] = self.signal_quote_off
        self.entry.signals['insert-toggle'] = self.signal_insert_toggle
        self.entry.signals['height-change'] = self.signal_height_change
        self.entry.signals['resize'] = self.signal_resize
    
        y, maxx = self.entry.get_status_line()
        self.statusbar = StatusBar(self.win, y, maxx)

        self.text = TextScroll(self.win, 0, y-1)
        self.text.redraw()
        self.text.refresh_signal = self.signal_refresh
        self.text.more_signal = self.signal_more

        self.update_status_line()
        self.entry.locate()

        self.action = {
            'reset': self.action_reset,
            'redraw': self.action_redraw,
            'macro': self.action_macro,
        }

        self.macro = { }

    def __del__(self):
        pass

    def add(self, line):
        self.text.add(line)

    def clear(self):
        self.text.clear_buffer()
        
    def set_more_mode(self, more_mode):
        self.text.set_more_mode(more_mode)
        if not more_mode:
            self.text.clear_more()

    def set_indent(self, n):
        self.text.wrap_prefix = n * ' '

    def set_word_wrap(self, n):
        self.text.word_wrap = n

    def set_status_line(self, val):
        self.status_line = val
        self.update_status_line()

    def set_expand(self, min, max):
        self.entry.set_expand(min, max)
        
    def update_status_line(self):
        table = self.status_table
        if self.quote_mode:
            table['quo'] = 'Quo'
        else:
            table['quo'] = '   '
        if self.insert_mode:
            table['ins'] = 'Ins'
            table['ovr'] = '   '
        else:
            table['ins'] = '   '
            table['ovr'] = 'Ovr'
        more_lines = self.text.get_more_lines()
        if more_lines > 0:
            table['more'] = '[MORE:%d]' % more_lines
        else:
            table['more'] = ''
        line = self.status_line % table
        self.statusbar.set_text(line)

    def draw(self):
        self.entry.locate()
        
    def reverse_lookup_key(self, key):
        if not self.entry.keymap.has_key(key):
            return ''
        func = self.entry.keymap[key]
        for i in self.entry.action.keys():
            if func == self.entry.action[i]:
                return i
        for i in self.action.keys():
            if func == self.action[i]:
                return i
        return '???'

    def show_key_binding(self, key, showing_all=0):
        name = self.reverse_lookup_key(key)
        k = key
        if ord(k[-1]) < 32:
            k = k[:-1] + '^' + chr(ord(k[-1])+64)
        elif ord(k[-1]) == 127:
            k = k[:-1] + '^?'
        elif k == ' ':
            k = 'space'
        if (name == 'macro') and self.macro.has_key(key):
            name = '"' + self.macro[key] + '"'
        # todo: action
        if showing_all:
            self.add('   %-16s%s' % (k, make_printable(name)))
        else:
            if len(name) > 0:
                self.add('*** Key binding: %s => %s' % (k, make_printable(name)))
            else:
                self.add('*** Key not bound.')

    def show_all_key_bindings(self):
        self.text.add('*** Key bindings:')
        list = self.entry.keymap.keys()
        list.sort()
        for k in list:
            self.show_key_binding(k, 1)

    def set_key_binding(self, key, action):
        func = None
        if self.action.has_key(action):
            func = self.action[action]
        elif self.entry.action.has_key(action):
            func = self.entry.action[action]
        elif action == '""':
            if self.entry.keymap.has_key(key):
                del self.entry.keymap[key]
            return -1
        elif (len(action) > 2) and (action[0] == '"') and (action[-1] == '"'):
            self.macro[key] = action[1:-1]
            func = self.action_macro
        else:
            return 0

        self.entry.keymap[key] = func
        return 1
        
    def action_reset(self, key):
        self.entry.reset()

    def action_redraw(self, key):
        # force window to be redrawn from scratch
        self.entry.window.clear()
        self.entry.window.refresh()
        self.entry.redraw()
        self.signal_resize(entry)
        self.entry.locate()

    def action_macro(self, key):
        if not self.macro.has_key(key):
            return
        for i in self.macro[key]:
            self.entry.insert_keypress(i)
#            self.entry.action_insert(i)

    def signal_quote_on(self, entry):
        self.quote_mode = 1
        self.update_status_line()

    def signal_quote_off(self, entry):
        self.quote_mode = 0
        self.update_status_line()

    def signal_insert_toggle(self, entry):
        self.insert_mode = entry.get_insert_mode()
        self.update_status_line()

    def signal_height_change(self, entry):
        y, maxx = entry.get_status_line()
        self.text.resize_bottom(y-1)
        self.statusbar.relocate(y, maxx)

    def signal_resize(self, entry):
        "unlike height_change, the width might have changed too"
        y, maxx = entry.get_status_line()
        self.text.resize_bottom(y-1)
        self.text.redraw()
        self.statusbar.relocate(y, maxx)

    def signal_refresh(self):
        self.entry.locate()

    def signal_more(self):
        self.update_status_line()
        self.entry.locate()
        
    ### dispatcher methods:

    def readable(self):
        return 1
    def writable(self):
        return 0
    def handle_connect(self):
        pass
    def handle_read(self):
        while self.entry.poll():
            if self.entry.is_complete():
                line = self.entry.get_text()
                self.entry.reset()
                self.text.clear_more()
                if hasattr(self, 'input_signal'):
                    self.active = not self.input_signal(line)
                self.update_status_line()
            self.entry.locate()
    def handle_close(self):
        self.active = 0

