#
# text entry widget for curses (in python, ported from perl).
# sort of a hybrid between the ircII/EPIC and TinyFugue text entries.
#
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# (c) 2000, 2001  Robey Pointer  <robey@lag.net>
#
#
# I coded this on the (long) night of 18 August 2000, with the help of 3
# keg-shaped Heineken cans.  Then in May 2001, I ported it to python because
# I was sick of perl.  It supports a lot of features:
# * the usual emacs control keys (C-a/b/d/e/f/k/l/n/p/t/u) by default,
#   as well as the equivalent PC keyboard keys (home, end, arrow keys, insert).
# * insert and overwrite modes (default C-o or insert to toggle) -- you can
#   attach a signal handler to modify your status bar when this mode changes
# * quote key (default C-v) allows the next key to be inserted as-is
# * completely configurable per-widget keymap, which can be used to hook your
#   app into special hotkeys, or also to remap keys to fit your aesthetic
# * 5 "meta-key" actions that can be assigned to keys in the keymap: these are
#   combined with the next keystroke to make a new keymap entry.  for example,
#   if you set C-x to action_meta1, then if the user hits "C-x n", it will
#   trigger the keymap entry for "meta1-n".
# * emits signals to notify you of a few things, like the screen being resized.
# * normally it sits at the bottom of the screen and takes up N lines.  you can
#   tell it to expand/shrink within a range of lines (like from 2 to 5 lines),
#   and it will start at the smallest size, expanding as the user types more
#   and more.  a signal is emitted whenever the widget grows or shrinks in this
#   way, so that you can resize other widgets.
# * keeps a history of past text entries in this widget, which can be viewed
#   and edited by the user.  you can have edited histories remembered in their
#   edited state (like bash) or be immutable (like ircii).
#
#
# signals:
#     quote-on (triggered when a quoted keypress begins)
#     quote-off (triggered when a quoted keypress ends)
#     insert-toggle (insert mode has changed state)
#     height-change (triggered when text entry shrinks or expands)
#     resize (triggered when a WINCH event occurs)


import curses, string

def putch(win, coords, ch):
    """
    display a character at a position in the window, using highlighting to
    mark characters that might otherwise be hidden (like control chars).
    """
    y, x = coords
    if ((ch >= ' ') and (ch <= '\x7E')) or ((ch >= '\xA1') and (ch <= '\xFF')):
        win.addstr(y, x, ch)
    elif ch < ' ':
        win.addstr(y, x, chr(ord(ch)+64), curses.A_REVERSE)
    elif ch == '\x7F':
        win.addstr(y, x, '?', curses.A_REVERSE)
    elif ch == '\xA0':
        win.addstr(y, x, ' ', curses.A_REVERSE)
    elif ch == '\x9F':
        win.addstr(y, x, '/', curses.A_REVERSE)
    else:
        # \x80 - \x9E
        win.addstr(y, x, chr(ord(ch)-32), curses.A_REVERSE)

# map curses key numbers to strings
Keymap = {
    curses.KEY_UP: 'up',
    curses.KEY_DOWN: 'down',
    curses.KEY_LEFT: 'left',
    curses.KEY_RIGHT: 'right',
    curses.KEY_HOME: 'home',
    curses.KEY_END: 'end',
    curses.KEY_FIND: 'home',	# ncurses bug
    curses.KEY_SELECT: 'end',	# ncurses bug
    curses.KEY_IC: 'insert',
    curses.KEY_DC: 'delete',
    curses.KEY_BACKSPACE: 'backspace',
    curses.KEY_RESIZE: 'resize',
    curses.KEY_F1: 'f1',
    curses.KEY_F2: 'f2',
    curses.KEY_F3: 'f3',
    curses.KEY_F4: 'f4',
    curses.KEY_F5: 'f5',
    curses.KEY_F6: 'f6',
    curses.KEY_F7: 'f7',
    curses.KEY_F8: 'f8',
    curses.KEY_F9: 'f9',
    curses.KEY_F10: 'f10',
    curses.KEY_F11: 'f11',
    curses.KEY_F12: 'f12',
}


class TextEntry:
# internal fields:
# height (# of rows to use)
# window (curses window to sit at the bottom of)
# text (input line so far)
# cursor (cursor's location, relative to zero)
# maxx, maxy (cached values of the window size)
# complete (set when the user has hit enter)
# insert_mode (chars should insert instead of overwriting)
# quote_mode (true when the next char is being quoted)
# meta_mode (1-5 if we're in a meta mode)
# expand_max (if set, let the text entry's height expand to N lines)
# expand_min (if set, start the text entry's height at N lines)
# history (list of previous entries)
# history_index (history entry in use)
# history_max (maximum number of items in the history list)
# history_buffer (temporary storage of text entry while browsing the history)
# history_keep_changes (if a history entry is edited but not entered, should those
#     changes be saved or tossed?)

    def __init__(self, window, height=1):
        self.height = height		# number of rows to use
        self.window = window		# curses window
        self.text = ''			# contents of text field
        self.cursor = 0			# cursor's position within text field
        self.insert_mode = 1		# insert vs. overwrite mode
        self.expand_max = 1		# maximum height of text field
        self.expand_min = 1		# minimum (starting) height of text field
        self.keymap = {}
        self.signals = {}
        self.history = []		# list of previous entries
        self.history_index = 0		# history entry currently being edited
        				#   (if history_index == len(history), editing a new field)
        self.history_max = 20		# maximum number of history entries to keep
        self.history_keep_changes = 1	# (see note below)
        self.complete = 0		# true when the user has hit [enter]
        self.quote_mode = 0		# true when the next key will be inserted as-is
        self.meta_mode = 0		# 1-5 when inside a meta-key sequence

        # history_keep_changes:
        # 1 - "bash mode": if you edit an entry in the history, and then leave
        #     without hitting enter, the edit is saved for later (you have
        #     changed the history)
        # 0 - "ircii mode": if you edit an entry in the history, and then leave
        #     without hitting enter, the edit is lost (the history remains
        #     immutable)

        # set the default keymap
        self.keymap = {
            '\x01': self.action_home,
            '\x02': self.action_left,
            '\x04': self.action_delete,
            '\x05': self.action_end,
            '\x06': self.action_right,
            '\x08': self.action_backspace,
            '\x0A': self.action_enter,
            '\x0B': self.action_deltoeol,
            '\x0C': self.action_redraw,
            '\x0D': self.action_enter,
            '\x0E': self.action_history_down,
            '\x0F': self.action_toggle_insert,
            '\x10': self.action_history_up,
            '\x14': self.action_transpose,
            '\x15': self.action_clear,
            '\x16': self.action_quote,
            '\x18': self.action_meta1,
            '\x7F': self.action_backspace,
            'backspace': self.action_backspace,
            'home': self.action_home,
            'end': self.action_end,
            'left': self.action_left,
            'right': self.action_right,
            'up': self.action_history_up,
            'down': self.action_history_down,
            'insert': self.action_toggle_insert,
            'delete': self.action_delete,
            'meta1-v': self.action_quote,
            'meta1-V': self.action_quote,
        }
        # also: f1-f12

        # convenient names for keymap actions
        self.action = {
            'self': self.action_insert,
            'home': self.action_home,
            'end': self.action_end,
            'left': self.action_left,
            'right': self.action_right,
            'enter': self.action_enter,
            'backspace': self.action_backspace,
            'delete': self.action_delete,
            'delete-eol': self.action_deltoeol,
            'clear-entry': self.action_clear,
            'redraw': self.action_redraw,
            'history-prev': self.action_history_up,
            'history-next': self.action_history_down,
            'insert-mode': self.action_toggle_insert,
            'quote-mode': self.action_quote,
            'transpose': self.action_transpose,
            'meta1': self.action_meta1,
            'meta2': self.action_meta2,
            'meta3': self.action_meta3,
            'meta4': self.action_meta4,
            'meta5': self.action_meta5,
        }

    def is_complete(self):
        "has the user hit return?  (you should check this after every poll)"
        return self.complete
    
    def check_size(self):
        "check the current window's size and cache it"
        self.maxy, self.maxx = self.window.getmaxyx()
        # hack to work around bug in python's curses library
        self.maxx = self.maxx - 1

    def get_status_line(self):
        "returns (y location of status line, maximum length of status line)"
        self.check_size()
        return (self.maxy - self.height - 1, self.maxx)

    def get_insert_mode(self):
        return self.insert_mode
    
    def clear(self):
        "clear the text entry's display area, without erasing the buffer"
        for i in range(self.maxy-self.height, self.maxy):
            self.window.move(i, 0)
            self.window.clrtoeol()

    def redraw(self):
        "redraw the widget, potentially taking into account any screen-size changes"
        self.check_size()
        self.clear()
        self.redraw_to_eol(0)
        self.window.move(self.cursor_coords())
        self.window.refresh()

    def notify_expand(self):
        if self.signals.has_key('height-change'):
            self.signals['height-change'](self)

    def reset(self):
        "clear text entry's contents and reset to minimum size"
        if self.height != self.expand_min:
            self.height = self.expand_min
            self.notify_expand()
        self.cursor = 0
        self.text = ''
        self.clear()
        self.window.refresh()
        self.complete = 0
        self.history_buffer = ''
        self.history_index = len(self.history)

    def check_expand(self):
        """
        called right before adding a character to the text entry's contents.
        if the text is filling up the max amount of space it can hold, and
        expand is on, and we still have room to expand, then we do so.
        """
        if (self.height == self.expand_max) or (len(self.text) < (self.maxx * self.height) - 1):
            return
        self.height = self.height + 1
        self.notify_expand()
        self.redraw()

    def set_text(self, text):
        "set the text entry's contents"
        expando = 0
        self.check_size()
        if len(text) >= self.maxx * self.expand_max:
            text = text[0:(self.maxx * self.expand_max)-1]
        while len(text) >= self.maxx * self.height:
            self.height = self.height + 1
            expando = 1
        if expando:
            self.notify_expand()

        self.text = text
        self.cursor = len(text)
        self.redraw()

    def get_text(self):
        "returns the text entry's contents"
        return self.text

    def set_expand(self, min, max):
        "set the text entry's minimum and maximum height"
        if min < 1:
            min = 1
        if max < min:
            max = min
        self.expand_min = min
        self.expand_max = max
        if min > self.height:
            self.height = min
            self.notify_expand()
            self.redraw()
        if max < self.height:
            self.height = max
            self.notify_expand()
            self.set_text(self.text)

    def get_expand(self):
        "returns the text entry's minimum and maximum height"
        return (self.expand_min, self.expand_max)

    def cursor_coords(self, n=None):
        "return (y, x) of where the cursor should be"
        if n == None:
            n = self.cursor
        return (self.maxy - self.height + (n / self.maxx), n % self.maxx)

    def redraw_to_eol(self, n=None):
        if n == None:
            n = self.cursor
        for i in range(n, len(self.text)):
            putch(self.window, self.cursor_coords(i), self.text[i])

    ##
    ##  functions that can be mapped to keys
    ##  all of these take one parameter: the name of the key pressed
    ##

    def action_backspace(self, key):
        if (len(self.text) == 0) or (self.cursor == 0):
            return
        self.cursor = self.cursor - 1
        self.text = self.text[0:self.cursor] + self.text[self.cursor+1:]
        self.redraw_to_eol()
        y, x = self.cursor_coords(len(self.text))
        self.window.addstr(y, x, ' ')

    def action_delete(self, key):
        if self.cursor == len(self.text):
            return
        self.cursor = self.cursor + 1
        self.action_backspace(key)

    # an emacs-ism.  we like those.
    def action_deltoeol(self, key):
        if self.cursor == len(self.text):
            return
        self.text = self.text[0:self.cursor]
        # FIXME: might be worth it to go ahead and manually redraw the deleted portions...
        self.redraw()

    def action_clear(self, key):
        self.text = ''
        self.cursor = 0
        self.redraw()

    def action_enter(self, key):
        self.complete = 1
        if len(self.text):
            self.history.append(self.text)
        if len(self.history) > self.history_max:
            self.history.pop(0)

    def action_insert(self, key):
        if self.insert_mode:
            self.check_expand()
            if len(self.text) >= (self.maxx * self.height) - 1:
                return
            self.text = self.text[0:self.cursor] + key + self.text[self.cursor:]
            self.redraw_to_eol()
        else:
            # overwrite mode
            if self.cursor == len(self.text):
                self.check_expand()
                if self.cursor >= (self.maxx * self.height) - 1:
                    return
            self.text = self.text[0:self.cursor] + key + self.text[self.cursor+1:]
            putch(self.window, self.cursor_coords(), key)
        self.cursor = self.cursor + 1

    def action_home(self, key):
        self.cursor = 0

    def action_end(self, key):
        self.cursor = len(self.text)

    def action_right(self, key):
        if self.cursor == len(self.text):
            return
        self.cursor = self.cursor + 1

    def action_left(self, key):
        if self.cursor == 0:
            return
        self.cursor = self.cursor - 1

    def action_transpose(self, key):
        if (self.cursor == 0) or (self.cursor >= len(self.text)):
            return
        left, right = self.text[self.cursor-1], self.text[self.cursor]
        self.text = self.text[:self.cursor-1] + right + left + self.text[self.cursor+1:]
        self.redraw()

    def action_redraw(self, key):
        self.redraw()
        if self.signals.has_key('resize'):
            self.signals['resize'](self)
 
    def action_toggle_insert(self, key):
        self.insert_mode = not self.insert_mode
        if self.signals.has_key('insert-toggle'):
            self.signals['insert-toggle'](self)

    def action_quote(self, key):
        self.quote_mode = 1
        self.window.keypad(0)
        if self.signals.has_key('quote-on'):
            self.signals['quote-on'](self)

    def action_history_up(self, key):
        if self.history_index == 0:
            return
        if self.history_index == len(self.history):
            # save current buffer
            self.history_buffer = self.get_text()
        elif self.history_keep_changes:
            # save modifications to current history entry
            self.history[self.history_index] = self.get_text()

        self.history_index = self.history_index - 1
        self.set_text(self.history[self.history_index])

    def action_history_down(self, key):
        if self.history_index == len(self.history):
            return
        if self.history_keep_changes:
            # save modifications to the current history entry
            self.history[self.history_index] = self.get_text()
        self.history_index = self.history_index + 1
        if self.history_index == len(self.history):
            self.set_text(self.history_buffer)
        else:
            self.set_text(self.history[self.history_index])

    def action_meta1(self, key):
        self.meta_mode = 1
    def action_meta2(self, key):
        self.meta_mode = 2
    def action_meta3(self, key):
        self.meta_mode = 3
    def action_meta4(self, key):
        self.meta_mode = 4
    def action_meta5(self, key):
        self.meta_mode = 5

    def keypress(self, ch):
        "interpret incoming curses keypress, and do an action"
        key = 'unknown'

        if ch > 255:
            # some kind of function key
            if Keymap.has_key(ch):
                key = Keymap[ch]
                if key == 'resize':
                    # special case!  resize event
                    self.redraw()
                    if self.signals.has_key('resize'):
                        self.signals['resize'](self)
                    return
            else:
                # unknown key; ignore
                return
        else:
            key = chr(ch)
            if self.quote_mode:
                self.action_insert(key)
                self.quote_mode = 0
                self.window.keypad(1)

                if self.signals.has_key('quote-off'):
                    self.signals['quote-off'](self)
                # move cursor back
                self.window.move(self.cursor_coords())
                self.window.refresh()
                return

        self.insert_keypress(key)

    def insert_keypress(self, key):
        if self.meta_mode:
            key = string.lower('meta' + str(self.meta_mode) + '-' + key)
            self.meta_mode = ''
            if not self.keymap.has_key(key):
                # bail before trying to self-insert
                return

        if self.keymap.has_key(key):
            self.keymap[key](key)
        else:
            # self-insert
            self.action_insert(key)

        # move cursor back
        self.window.move(self.cursor_coords())
        self.window.refresh()

    def poll(self):
        self.window.move(self.cursor_coords())
        self.window.refresh()

        ch = self.window.getch()
        if ch != -1:
            self.keypress(ch)
            return 1
        return 0

    def locate(self):
        self.window.move(self.cursor_coords())
        self.window.refresh()
