#
# ported from perl
#

import curses, re

format_re = re.compile('\006([^\006]*)\006');

# weird hack
BREAK = '\006XB\006'

def putline(window, y, str, attrs={}):
    """
    draw a formatted string 'str' in 'window' on line 'y'.
    'attrs' is any attributes saved from a previous call.
    returns: an attribute object suitable for using in future calls, to
        preserve attributes (for example, on wrapped lines)
    """

    # in the perl version, this 'attrs' thing was just a string that was
    # dumped thru regex's to get the attributes (a string like "B4U2").
    # this seems more reasonable, even tho it might take up more memory.
    bold = attrs.get('bold', 0)
    reverse = attrs.get('reverse', 0)
    under = attrs.get('under', 0)
    italics = attrs.get('italics', 0)
    curses_attr = 0

    window.move(y, 0)
    window.clrtoeol()
    i, x = 0, 0
    while i < len(str):
        if str[i] == '\006':
            # special formatting
            m = format_re.search(str[i:])
            if m == None:
                cmd = ''
            else:
                cmd = m.group(1)
            if cmd == 'B-':
                if bold > 0: bold = bold - 1
            elif (cmd == 'B') or (cmd == 'B+'):
                bold = bold + 1
            elif cmd == 'V-':
                if reverse > 0: reverse = reverse - 1
            elif (cmd == 'V') or (cmd == 'V+'):
                reverse = reverse + 1
            elif cmd == 'U-':
                if under > 0: under = under - 1
            elif (cmd == 'U') or (cmd == 'U+'):
                under = under + 1
            elif cmd == 'I-':
                if italics > 0: italics = italics - 1
            elif (cmd == 'I') or (cmd == 'I+'):
                italics = italics + 1
            elif cmd == 'N':
                italics = under = reverse = bold = 0
            elif cmd == 'XB':
                # refresh attributes -- just being here is enough, because
                # it'll fall thru to the refreshing of curses_attr below
                pass
            else:
                pass
            curses_attr = 0
            if bold: curses_attr = curses_attr | curses.A_BOLD
            if reverse: curses_attr = curses_attr | curses.A_REVERSE
            if under: curses_attr = curses_attr | curses.A_UNDERLINE
            if italics: curses_attr = curses_attr | curses.A_DIM
            i = i + len(cmd) + 1
        else:
            try:
                window.addstr(y, x, str[i], curses_attr)
            except curses.error:
                pass	# i don't give a shit.
            x = x + 1
        i = i + 1

    attrs['bold'] = bold
    attrs['reverse'] = reverse
    attrs['under'] = under
    attrs['italics'] = italics
    return attrs

def count_oob(str):
    "count the number of chars in a string that are used for ^F formatting chunks"
    count = 0
    while 1:
        m = format_re.search(str)
        if m == None:
            break
        count = count + 2 + len(m.group(1))
        str = format_re.sub('', str, 1)
    return count

def str_find(str, n):
    "find the N'th char of a string, ignoring format chars, and return the actual N"
    i, i_real = 0, 0
    in_format = 0
    while (i < n) and (i_real < len(str)):
        if str[i_real] == '\006':
            in_format = not in_format
        elif not in_format:
            i = i + 1
        i_real = i_real + 1
    return i_real


class TextScroll:

    def __init__(self, window, top=0, bottom=23):
        self.window = window		# curses window
        self.top = top			# top row to write on (starting at 0)
        self.bottom = bottom		# bottom row to write on
        self.width = 80			# width of window (discovered via curses)
        self.text = [ ]			# list of text lines in the display
        self.buffer = [ ]		# list of text lines buffered for "more" mode
        self.wrap = 1			# 1: wrap long lines into multiple lines
        				# 0: truncate long lines at the right edge
        self.word_wrap = 1		# if true, wrap only at word boundaries
        self.wrap_prefix = '    '	# text to insert before wrapped lines
        self.scrollback = 100		# maximum number of lines to save
        self.top_mark = -1		# first unacknowledged line of the text buffer*
        self.holding = 0		# 1: currently paused on "more"
        				# 2: paused on "more" in the middle of a wrapped line
        self.more_signal = ''		# function to call when "more" mode triggers
        self.refresh_signal = ''	# function to call when the screen needs updating

        # * in more mode, this is the line that will scroll up to be the top
        #   of the screen (and then we'll buffer everything after that).
        #   otherwise, it's -1.

        self.check_width()

    def check_width(self):
        None, self.width = self.window.getmaxyx()
        # hack to get around a bug in python's curses
        self.width = self.width - 1

    def clear(self):
        "clear visible area, without modifying the cached text to be displayed"
        for i in range(self.top, self.bottom + 1):
            self.window.move(i, 0)
            self.window.clrtoeol()

    def break_line(self, text):
        """
        break a text line into as many lines as it would take to fit into this
        screen width (following word-wrap, etc)
        """
        if not self.wrap:
            # truncate and return
            return [ text[:str_find(text, self.width)] ]

        if not self.word_wrap:
            out = [ ]
            # just chop up at exact word boundaries
            while len(text) - count_oob(text) >= self.width:
                l = str_find(text, self.width-1)
                out.append(text[0:l])
                text = self.wrap_prefix + BREAK + text[l:]
            out.append(text)
            return out

        # word wrap if necessary
        min = 0
        out = [ ]
        while len(text) - count_oob(text) > self.width:
            n = str_find(text, self.width-1)
            while (text[n] != ' ') and (text[n] != '-') and (n > min):
                n = n - 1
            # if we didn't find any line-break chars, do a force break.
            if n == min:
                n = str_find(text, self.width-1)
            out.append(text[0:n+1])
            text = self.wrap_prefix + BREAK + text[n+1:]
            min = len(self.wrap_prefix)
        out.append(text);
        return out

    def start_more(self):
        "turn on more mode"
        if not self.holding:
            self.holding = 1
            if self.more_signal != None:
                self.more_signal()

    def build_current(self):
        """
        assemble a list of what should be showing on the screen currently.
        returns: list of logical lines, where each logical line is a list of
            physical lines.
        """
        screen = [ ]
        total = 0
        max = self.bottom - self.top + 1
        top_mark = self.top_mark
        if top_mark < 0:
            top_mark = len(self.text)-1
	    if top_mark < 0:
		top_mark = 0
        i = top_mark

        # first, work from top_mark down to the bottom of the buffer
        while (total < max) and (i < len(self.text)):
            list = self.break_line(self.text[i])
            total = total + len(list)
            while total > max:
                if self.top_mark < 0:
                    # not in more mode, so cut off the top
                    list.pop(0)
                else:
                    # cut off the bottom
                    self.start_more()
                    list.pop()
                total = total - 1
            screen.append(list)
            i = i + 1

        if i < len(self.text):
            # put extra lines into the more buffer.
            # (this can happen if the user resizes the screen, suddenly
            # creating the need to go into more-mode)
            self.start_more()
            while i < len(self.text):
                self.buffer.insert(0, self.text.pop())
            return screen

        # now, work backwards from that mark
        i = top_mark - 1
        while (total < max) and (i >= 0):
            list = self.break_line(self.text[i])
            total = total + len(list)
            while total > max:
                list.pop(0)
                total = total - 1
            screen.insert(0, list)
            i = i - 1

        # fill in blank lines
        while total < max:
            screen.insert(0, [''])
            total = total + 1

        return screen

    def redraw(self):
        "redraw the window"
        self.check_width()
        screen = self.build_current()
        line = self.top
        while line <= self.bottom:
            list = screen.pop(0)
            attr = {}
            for str in list:
                attr = putline(self.window, line, str, attr)
                line = line + 1
        self.window.refresh()

    def add(self, text):
        "add content to the bottom of the display"
        if self.holding:
            # just buffer things while waiting for "more"
            self.buffer.append(text)
            return

        list = self.break_line(text)

        if 0:
            # debug
            y = 2
            for i in list:
                self.window.addstr(y, 0, i)
                y = y + 1
            self.window.addstr(y, 0, '---')

        if self.top_mark >= 0:
            lines_left = self.bottom - self.top + 1
            for i in range(self.top_mark, len(self.text)):
                sublist = self.break_line(self.text[i])
                lines_left = lines_left - len(sublist)
            if lines_left == 0:
                # perfect fill, amazing!
                self.start_more()
                self.buffer.append(text)
                return
            # it's possible that lines_left < 0 here :(
            # this is probably not the right answer.  but i don't know what is.
            if lines_left < 0:
                self.start_more()
                self.buffer.append(text)
                return
            while lines_left < len(list):
                self.start_more()
                list.pop()
                # make secret note to self that i displayed a half line:
                self.holding = 2

        while len(self.text) >= self.scrollback:
            self.text.pop(0)
            if self.top_mark >= 0:
                self.top_mark = self.top_mark - 1
        self.text.append(text)

        # scroll N lines up
        self.window.idlok(1)
        self.window.scrollok(1)
        self.window.setscrreg(self.top, self.bottom)
        self.window.scroll(len(list))
        self.window.scrollok(0)

        attr = {}
        i, y = 0, self.bottom
        if y-(len(list)-1) < self.top:
            i = (len(list)-1) - y
        while i < len(list):
            attr = putline(self.window, y-(len(list)-1)+i, list[i], attr)
            i = i + 1
        if self.refresh_signal:
            self.refresh_signal()
#        self.window.refresh()

    def clear_more(self):
        "reset the more count (usually because the user hit enter or something)"
        if not self.holding:
            self.top_mark = len(self.text)
            return
        buffer = self.buffer
        self.buffer = [ ]
        self.top_mark = len(self.text)
        if self.holding == 2:
            # we were holding on a half-line
            self.top_mark = self.top_mark - 1
            self.redraw()
        self.holding = 0
        # dump buffer into the scroll region, as if it was just typed.
        # any overflow will return to the buffer all over again.
        while len(buffer):
            self.add(buffer.pop(0))

    def clear_buffer(self):
        "wipe out any pending more-mode buffer"
        self.buffer = [ ]
        self.holding = 0
        self.top_mark = 0
        self.clear()
        self.text = [ ]

    def resize_bottom(self, bottom):
        "change the bottom edge of the text window, and possibly resize"
        if bottom < self.bottom:
            # just scroll up
	    try:
                self.window.idlok(1)
                self.window.scrollok(1)
                self.window.setscrreg(self.top, self.bottom)
                self.window.scroll(self.bottom - bottom)
	    except curses.error:
		pass
            self.window.scrollok(0)
            self.bottom = bottom
        else:
            # could do something clever, but easiest just to redraw
            self.bottom = bottom
            self.redraw()

    def set_more_mode(self, mode):
        "turn on/off the ability to be in more-mode"
        if mode:
            if self.top_mark < 0:
                self.top_mark = len(self.text)-1
            if self.top_mark < 0:
                self.top_mark = 0
        else:
            self.top_mark = -1
            if self.holding:
                self.holding = 0
                self.redraw()

    def get_more_lines(self):
        "returns number of lines buffered in more-mode"
        return len(self.buffer)

    def debug(self):
        return "(top:%d,txt:%d,buf:%d,%d)" % self.top_mark, len(self.text), len(self.buffer), self.holding
