# python bindings for enaml
# started 17 mar 2001, completed 18 mar 2001
#
# (c) 2001  Robey Pointer <robey@lag.net>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

from types import *
import string

# how deep we'll traverse nested blocks or lists before throwing in the towel
_MAX_DEPTH = 32

def _iskey(x):
    return (x >= 'a' and x <= 'z') or (x >= 'A' and x <= 'Z') or x == '-' or x == '_'

def _isdigit(x):
    return x >= '0' and x <= '9'
    
def _iswhitespace(x):
    return x == ' ' or x == '\t' or x == '\x0D' or x == '\x0A'

def _ishex(x):
    return (x >= 'a' and x <= 'f') or (x >= 'A' and x <= 'F') or _isdigit(x)

def _needsquote(x):
    return x < ' ' or x == '\x7F' or x == '\xFF' or x == '"' or x == '%'

def _unhex(x):
    if (x >= 'A' and x <= 'F'):
        return ord(x) - ord('A') + 10
    elif (x >= 'a' and x <= 'f'):
        return ord(x) - ord('a') + 10
    elif (x >= '0' and x <= '9'):
        return ord(x) - ord('0')
    else:
        return 0

def _unhtmlify(str):
    "convert embedded %HH in a string into a byte ('%34' => '4')"
    out = ''
    i = 0
    while i < len(str):
        if str[i] == '%' and i+2 < len(str):
            out = out + chr(_unhex(str[i+1]) * 16 + _unhex(str[i+2]))
            i = i + 2
        else:
            out = out + str[i]
        i = i + 1
    return out

def _unhexify(str):
    "convert string of hex digits into a string of bytes"
    out = ''
    for i in range(0, len(str), 2):
        if i+1 >= len(str):
            continue
        out = out + chr(_unhex(str[i]) * 16 + _unhex(str[i+1]))
    return out

_TOKEN_EOF, _TOKEN_UNKNOWN, _TOKEN_OPEN_BRACKET, _TOKEN_CLOSE_BRACKET, _TOKEN_OPEN_BRACE, \
            _TOKEN_CLOSE_BRACE, _TOKEN_COLON, _TOKEN_HEX, _TOKEN_STRING, _TOKEN_NUMBER, \
            _TOKEN_KEY = range(-1, 10)

def _getToken(str):
    "returns (token-type, token-string, remainder-string)"

    while len(str) > 0 and _iswhitespace(str[0]):
        str = str[1:]
    if len(str) == 0:
        return _TOKEN_EOF, '', ''
        
    if str[0] == '[':
        return _TOKEN_OPEN_BRACKET, str[0], str[1:]
    elif str[0] == ']':
        return _TOKEN_CLOSE_BRACKET, str[0], str[1:]
    elif str[0] == '{':
        return _TOKEN_OPEN_BRACE, str[0], str[1:]
    elif str[0] == '}':
        return _TOKEN_CLOSE_BRACE, str[0], str[1:]
    elif str[0] == ':':
        return _TOKEN_COLON, str[0], str[1:]

    if str[0] == '%':
        left, right = 1, 1
        while right < len(str) and _ishex(str[right]):
            right = right + 1
        return _TOKEN_HEX, _unhexify(str[left:right]), str[right:]

    if str[0] == '"' or str[0] == "'":
        quote = str[0]
        left, right = 1, 1
        while right < len(str) and str[right] != quote:
            right = right + 1
        if right < len(str):
            # skip trailing quote
            return _TOKEN_STRING, _unhtmlify(str[left:right]), str[right+1:]
        else:
            return _TOKEN_STRING, _unhtmlify(str[left:right]), str[right:]

    if _isdigit(str[0]):
        right = 0
        while right < len(str) and _isdigit(str[right]):
            right = right + 1
        return _TOKEN_NUMBER, int(str[0:right]), str[right:]

    if _iskey(str[0]):
        right = 0
        while right < len(str) and _iskey(str[right]):
            right = right + 1
        return _TOKEN_KEY, str[0:right], str[right:]
    
    return _TOKEN_UNKNOWN, str[0], str[1:]

def parse(str):
    """parse(str) -> dictionary

    Returns a dictionary parsed out of an Enaml protocol string.
    If the string is malformed, None is returned.

    The opposite of 'parse' is 'compose'."""

    # closest python comes to enum
    STATE_BLOCK, STATE_LIST = range(2)
    
    state = STATE_BLOCK
    block = {}		# block being parsed
    list = None		# list being parsed
    block_stack = []	# previous ancestor blocks
    list_stack = []	# previous ancestor lists
    key = ''		# current key
    
    while 1:
        token, value, str = _getToken(str)
        if token == _TOKEN_OPEN_BRACE:
            if not key or state != STATE_BLOCK or len(block_stack) >= _MAX_DEPTH:
                return None
            block_stack.append(block)
            block = {}
            block_stack[len(block_stack)-1][key] = block
            key = ''
        elif token == _TOKEN_CLOSE_BRACE:
            if len(block_stack) == 0 or state != STATE_BLOCK:
                return None
            if key:
                block[key] = None   	# flag
                key = ''
            block = block_stack.pop()
        elif token == _TOKEN_OPEN_BRACKET:
            if state == STATE_BLOCK:
                if not key:
                    return None
                list = []
                block[key] = list
                state = STATE_LIST
                key = ''
            else:
                if (len(list_stack) >= _MAX_DEPTH):
                    return None
                list_stack.append(list)
                list = []
                list_stack[len(list_stack)-1].append(list)
        elif token == _TOKEN_CLOSE_BRACKET:
            if state != STATE_LIST:
                return None
            if len(list_stack) > 0:
                list = list_stack.pop()
            else:
                list = None
                state = STATE_BLOCK
        elif token == _TOKEN_COLON:
            if state != STATE_BLOCK or not key:
                return None
        elif token == _TOKEN_NUMBER or token == _TOKEN_STRING or token == _TOKEN_HEX:
            if state == STATE_BLOCK:
                if not key:
                    return None
                block[key] = value
                key = ''
            else:
                list.append(value)
        elif token == _TOKEN_KEY:
            if state == STATE_LIST:
                return None
            if key:
                block[key] = None	# flag
            key = string.lower(value);
        elif token == _TOKEN_UNKNOWN:
            return None
        elif token == _TOKEN_EOF:
            # clean up and go home
            if key:
                block[key] = None	# flag
            if len(block_stack) > 0:
                block = block_stack[0]
            return block

def _hexquote(str):
    out = ''
    for c in str:
        out = out + ('%02X' % ord(c))
    return out

def _quote(str):
    quote_count = 0
    for c in str:
        if (_needsquote(c)):
            quote_count = quote_count + 1
    if quote_count > len(str)/2:
        # might as well use a hex-quoted string
        return '%' + _hexquote(str)

    out = '"'
    for c in str:
        if (_needsquote(c)):
            out = out + '%' + _hexquote(c)
        else:
            out = out + c
    return out + '"'

def _toString(value):
    "convert a python value into an enaml string"
    if type(value) is DictionaryType:
        return '{ ' + _table_compose(value) + '}'
    elif type(value) is ListType:
        return '[ ' + _list_compose(value) + ']'
    elif type(value) is IntType:
        return str(value)
    else:
        return _quote(str(value))

def _list_compose(list):
    out = ''
    for item in list:
        out = out + _toString(item) + ' '
    return out

def _table_compose(table):
    out = ''
    for key in table.keys():
        value = table[key]
        if type(value) is NoneType:
            out = out + str(key) + ' '
        else:
            out = out + str(key) + ':' + _toString(value) + ' '
    return out

def compose(key, table=None):
    """compose(key, table=None) -> string

    Return an Enaml string composed of the initial key and a dictionary.
    Example: compose('status', {'message': 'Going home!'})
        -> 'status:{ message:"Going home!" }'
    Example: compose('ping')
        -> 'ping'
    Example: compose('icon-change', {'icon': '\x47\x49\x46\x38\x39\x61\xFF'})
        -> 'icon-change:{ icon:"GIF89a%FF" }'

    The opposite of 'compose' is 'parse'."""
    
    if not (type(table) is DictionaryType):
        return str(key)
    if not key:
        return '{ ' + _table_compose(table) + '}'
    else:
        return str(key) + ':{ ' + _table_compose(table) + '}'

def get(table, key, default=''):
    """get(table, key, default='') -> string

    Fetch a string out of a table, if it exists.
    If it doesn't exist, return the default.
    """

    if table.has_key(key):
        return str(table[key])
    else:
        return default


    
