# a python class to connect to an enaml-based server (for example, say2) and
# read/write protocol messages easily.  asyncore made this a trivial exercise.
# 19 March 2001
#
# (c) 2001  Robey Pointer <robey@lag.net>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

import sys, string
import enaml
import socket
import asyncore

# this never stays the same in asyncore
def compact_traceback():
    t,v,tb = sys.exc_info()
    tbinfo = []
    while 1:
        tbinfo.append((tb.tb_frame.f_code.co_filename, tb.tb_frame.f_code.co_name, str(tb.tb_lineno)))
        tb = tb.tb_next
        if not tb:
            break

    file, function, line = tbinfo[-1]
    info = str(t) + ':' + str(v) + ' [' + string.join(map(lambda x: string.join(x, '|'), tbinfo), '] [') + ']'
    return (file, function, line), info

class EnamlClient(asyncore.dispatcher):

    """
    Speak to a server that talks in Enaml (for example, a say2 server).
    http://www.lag.net/say2/enaml.txt
    
    methods:
    connect(host, port): connect to enaml server
    disconnect(): er, uh, disconnect
    recv_message() -> table or None: return next enaml message
    is_message_queued(): return T if there are enaml message(s) waiting
    send_message(key, table): send enaml message

    important:
    asyncore.poll(timeout): trigger read/write callbacks
    don't override handle_connect!  (need it to work around a bug in asyncore)
    
    signals:
    message_signal(self): at least one new protocol message has arrived
    handle_connect_success(self): connection succeeded
    handle_connect_failed(self, errno): connection failed
    handle_close(self): [inherited] connection is closed
    log(self, string): [inherited] log socket events (somewhat annoying)
    """
    
    def __init__(self, host=None, port=None):
        asyncore.dispatcher.__init__(self)
        self.maybe_connect = 0
        if (host != None) and (port != None):
            self.connect(host, port)

    def connect(self, host, port):
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        asyncore.dispatcher.connect(self, (host, port))
        self.host = host
        self.port = port
        # not sure if this is all necessary but i'm paranoid
        self.in_buffer = ''
        self.in_buffer_size = 8192
        self.out_buffer = ''
        self.message_queue = []

    def disconnect(self):
        self.handle_close()
        
    def handle_connect(self):
        # there's a bug in asyncore: after starting a nonblocking connect(),
        # when the socket triggers for read access (which indicates that you
        # should do a read() to get your status back), asyncore assumes
        # success and calls handle_connect().  therefore we need to flag this
        # possibility and catch it later.
        # THIS MEANS YOU SHOULD NOT OVERRIDE handle_connect().
        self.maybe_connect = 1

    def handle_connect_success(self):
        pass
    
    def handle_connect_failed(self, why):
        pass

    def handle_close(self):
        self.close()
        
    def handle_read(self):
        # refill buffer
        try:
            self.in_buffer = self.in_buffer + self.recv(self.in_buffer_size - len(self.in_buffer))
        except socket.error, why:
            if self.maybe_connect:
                # because we're nonblocking, this is how we'll find out if a connection failed.
                self.handle_connect_failed(why)
                self.maybe_connect = 0
                self.close()
                return
            else:
                raise socket.error, why
        if self.maybe_connect:
            self.handle_connect_success()
            self.maybe_connect = 0

        # is it possible to pull of a complete line?
        if len(self.in_buffer) == self.in_buffer_size and not ('\n' in self.in_buffer):
            # pretend there's an implicit linefeed at the end of the buffer
            block = enaml.parse(self.in_buffer)
            self.in_buffer = ''
            if block != None:
                self.message_queue.append(block)
                self.message_signal()
            return
        signal = 0
        while '\n' in self.in_buffer:
            # python makes it really painful to do this :(
            i = string.find(self.in_buffer, '\n')
            if i > 0 and self.in_buffer[i-1] == '\r':
                line = self.in_buffer[:i-1]
            else:
                line = self.in_buffer[:i]
            self.in_buffer = self.in_buffer[i+1:]
            if len(line) > 0:
                block = enaml.parse(line)
                if block != None:
                    self.message_queue.append(block)
                    signal = 1
        # only send the signal once, they'll get the idea
        if signal:
            self.message_signal()

    def handle_write(self):
        sent = self.send(self.out_buffer)
        self.out_buffer = self.out_buffer[sent:]

    def writable(self):
        return (len(self.out_buffer) > 0)

    def readable(self):
        return (len(self.in_buffer) < self.in_buffer_size)

    # METHODS SPECIFIC TO ENAML_CLIENT

    def recv_message(self):
        "recv_message(self) -> table (or None)"
        if len(self.message_queue) > 0:
            return self.message_queue.pop(0)
        return None

    def is_message_queued(self):
        return (len(self.message_queue) > 0)
    
    def send_message(self, key, block=None):
        try:
            self.send(enaml.compose(key, block) + '\r\n')
        except socket.error, why:
            t, info = compact_traceback()
            self.log('exception: ' + why[1] + ' at ' + info)
            
    # DEFAULT SIGNAL HANDLERS

    def message_signal(self):
        while self.is_message_queued():
            print self.recv_message()

