# a python class for talking to a say2 server.
# 19 March 2001
#
# (c) 2001  Robey Pointer <robey@lag.net>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

import enaml
import asyncore
import enaml_client

class Say2Client(enaml_client.EnamlClient):
    """
    Wrapper for enaml_client that handles pings automatically, and maps other
    messages into callbacks.
    """

    def __init__(self, host, port=3797):
        enaml_client.EnamlClient.__init__(self, host, port)
        self.listeners = []

    def log(self, string):
        self.event('log', string)

    def log_info(self, message, type='info'):
        if type != 'info':
            self.event('log', '%s: %s' % (type, message))

    def handle_close(self):
        self.event('close')
        enaml_client.EnamlClient.handle_close(self)

    def handle_connect_success(self):
        self.event('connect')
        
    def handle_connect_failed(self, why):
        self.event('connect_failed', why)

    def add_listener(self, hook):
        if not hook in self.listeners:
            self.listeners.append(hook)

    def remove_listener(self, hook):
        if hook in self.listeners:
            self.listeners.remove(hook)

    def login(self, login, password, client_name='anonymous python client', replicant=0, address=None):
        block = { 'id': login, 'password': password, 'client': client_name }
        if replicant:
            block['replicant'] = None
        if address != None:
            block['address'] = str(address[0]) + ':' + str(address[1])
        self.send_message('login', block)

    def login_guest(self, login, email, tagline=None, client_name='anonymous python client', address=None):
        block = { 'nick': login, 'email': email, 'client': client_name }
        if tagline != None:
            block['tagline'] = tagline
        if address != None:
            block['address'] = str(address[0]) + ':' + str(address[1])
        self.send_message('login', block)

    def logout(self):
        self.send_message('logout')
        
    def set_tagline(self, tagline):
        self.send_message('prefs', { 'tagline': tagline })

    def set_password(self, password):
        self.send_message('prefs', { 'password': password })

    def set_icon(self, icon):
        self.send_message('prefs', { 'icon': icon })

    def set_status(self, status, text=''):
        self.send_message('status', { 'mode': status, 'text': text })

    def set_nick(self, nick):
        self.send_message('change-nick', { 'nick': nick })

    def whois(self, nick, id):
        if id != None:
            self.send_message('whois', { 'id': id })
        if nick != None:
            self.send_message('whois', { 'nick': nick })
            
    def list_users(self, verbose=0, grep=None):
        table = { 'type': 'user' }
        if grep != None:
            table['grep'] = grep
        if verbose:
            table['verbose'] = None
        self.send_message('list-clients', table)

    def list_channels(self, grep=None):
        if grep == None:
            self.send_message('list-clients', { 'type': 'channel' })
        else:
            self.send_message('list-clients', { 'type': 'channel', 'grep': grep })

    def get_server_cert(self):
        self.send_message('server-cert')

    def say(self, text):
        self.send_message('say', { 'text': text })

    def say_act(self, action):
        self.send_message('say', { 'act': action })

    def wall(self, text):
        self.send_message('wall', { 'text': text })
        
    def message(self, nick, id, text):
        if id != None:
            self.send_message('message', { 'id': id, 'text': text })
        elif nick != None:
            self.send_message('message', { 'nick': nick, 'text': text })

    def message_act(self, nick, id, action):
        if id != None:
            self.send_message('message', { 'id': id, 'act': action })
        elif nick != None:
            self.send_message('message', { 'nick': nick, 'act': action })

    def boot(self, nick, id, text):
        block = { }
        if id != None:
            block['id'] = id
        elif nick != None:
            block['nick'] = nick
        if text != None:
            block['text'] = text
        self.send_message('boot-user', block)

    def get_history(self, max=0):
        if max > 0:
            self.send_message('history', { 'max': str(max) })
        else:
            self.send_message('history')

    def subscribe(self, id, text=None):
        block = { 'id': id }
        if text != None:
            block['text'] = text;
        self.send_message('subscribe', block)

    def unsubscribe(self, id):
        self.send_message('unsubscribe', { 'id': id })
        
    def subscription_accept(self, id, text=None):
        block = { 'id': id }
        if text != None:
            block['text'] = text;
        self.send_message('subscription-accepted', block)

    def subscription_deny(self, id, text=None):
        block = { 'id': id }
        if text != None:
            block['text'] = text;
        self.send_message('subscription-denied', block)

    def go(self):
        self.got_something = 1
        while self.got_something:
            self.got_something = 0
            asyncore.poll(0.25)

    def event(self, event, *params):
        # kinda weird: for each listener, we see if they implement a method
        # "event_<event>", and if so, call it.
        if not hasattr(self, 'listeners'):
            # might be called before constructor is done
            return
        for obj in self.listeners:
            if hasattr(obj.__class__, 'event_' + event):
                try:
                    apply(getattr(obj, 'event_' + event), params)
                except:
                    t, info = enaml_client.compact_traceback()
                    self.log('EXCEPTION: ' + info)

    def message_signal(self):
        while self.is_message_queued():
            block = self.recv_message()
            if block == None:
                return
            for command in block.keys():
                sub = block[command]
                if sub == None:
                    sub = {}
                self.got_something = 1

                if command == 'ping':
                    self.send_message('ping')
                elif command == 'hello':
                    if sub.has_key('protocol'):
                        version = sub['protocol']
                    else:
                        version = [0, 0]
                    self.event('hello', version)
                elif command == 'error':
                    command = enaml.get(sub, 'command', '')
                    text = enaml.get(sub, 'text', '')
                    self.event('error', command, text)
                elif command == 'not-present':
                    nick = enaml.get(sub, 'nick', '')
                    id = enaml.get(sub, 'id', '')
                    self.event('not_present', nick, id)
                elif command == 'server-info':
                    server = enaml.get(sub, 'server', 'unknown')
                    version = enaml.get(sub, 'version', '')
                    text = enaml.get(sub, 'text', '')
                    uptime = int(enaml.get(sub, 'uptime', '0'))
                    self.event('server_info', server, version, text, uptime)
                elif command == 'welcome':
                    nick = enaml.get(sub, 'nick', '')
                    id = enaml.get(sub, 'id', '')
                    attached = sub.has_key('attached')
                    replicant = sub.has_key('replicant')
                    host = enaml.get(sub, 'host', '')
                    self.nick, self.id = nick, id
                    self.event('welcome', nick, id, attached, replicant, host)
                elif command == 'replicant-terminated':
                    host = enaml.get(sub, 'host', '')
                    self.event('replicant_terminated', host)
                elif command == 'bad-nick':
                    nick = enaml.get(sub, 'nick', '')
                    text = enaml.get(sub, 'text', '')
                    self.event('bad_nick', nick, text)
                elif command == 'bad-address':
                    address = enaml.get(sub, 'address', '')
                    self.event('bad_address', address)
                elif command == 'no-such-client':
                    id = enaml.get(sub, 'id', '')
                    typ = enaml.get(sub, 'type', 'user')
                    self.event('no_such_client', typ, id)
                elif command == 'logout':
                    text = enaml.get(sub, 'text', '')
                    detached = enaml.get(sub, 'detached', '')
                    self.event('logout', text, detached)
                elif command == 'motd':
                    text = enaml.get(sub, 'text', None)
                    if text != None:
                        self.event('motd', text)
                elif command == 'signon':
                    nick = enaml.get(sub, 'nick', '')
                    id = enaml.get(sub, 'id', '')
                    time = int(enaml.get(sub, 'time', '0'))
                    email = enaml.get(sub, 'email', '')
                    self.event('signon', nick, id, time, email)
                elif command == 'signoff':
                    nick = enaml.get(sub, 'nick', '')
                    id = enaml.get(sub, 'id', '')
                    time = int(enaml.get(sub, 'time', '0'))
                    self.event('signoff', nick, id, time)
                elif command == 'prefs':
                    tagline = enaml.get(sub, 'tagline', '')
                    # notify is going away
                    notify = enaml.get(sub, 'notify', '')
                    pwchanged, iconchanged = 0, 0
                    if sub.has_key('password-changed'):
                        pwchanged = 1
                    if sub.has_key('icon'):
                        iconchanged = 1
                    self.event('prefs', tagline, notify, pwchanged, iconchanged)
                elif command == 'status':
                    nick = enaml.get(sub, 'nick', '')
                    id = enaml.get(sub, 'id', '')
                    time = int(enaml.get(sub, 'time', '0'))
                    mode = enaml.get(sub, 'mode', '')
                    text = enaml.get(sub, 'text', '')
                    self.event('status', nick, id, time, mode, text)
                elif command == 'change-nick':
                    id = enaml.get(sub, 'id', '')
                    old_nick = enaml.get(sub, 'old-nick', '')
                    nick = enaml.get(sub, 'nick', '')
                    self.event('change_nick', old_nick, id, nick)
                elif command == 'members':
                    typ = enaml.get(sub, 'type', 'user')
                    if sub.has_key('list') and type(sub['list']) == type([]):
                        list = sub['list']
                    else:
                        list = []
                    self.event('members', typ, list)
                elif command == 'server-cert':
                    host = enaml.get(sub, 'host', '')
                    cert = enaml.get(sub, 'cert', None)
                    self.event('server_cert', host, cert)
                elif command == 'client-info':
                    # hoo boy.  just send the whole thing
                    # (but make sure all the required fields are there)
                    if not sub.has_key('id'): sub['id'] = ''
                    if not sub.has_key('type'): sub['type'] = 'user'
                    if not sub.has_key('email'): sub['type'] = ''
                    self.event('client_info', sub)
                elif command == 'say':
                    nick = enaml.get(sub, 'nick', '')
                    id = enaml.get(sub, 'id', '')
                    time = int(enaml.get(sub, 'time', '0'))
                    self.event('say', nick, id, time, sub)
                elif command == 'wall':
                    nick = enaml.get(sub, 'nick', '')
                    id = enaml.get(sub, 'id', '')
                    time = int(enaml.get(sub, 'time', '0'))
                    self.event('wall', nick, id, time, sub)
                elif command == 'message':
                    nick = enaml.get(sub, 'nick', '')
                    id = enaml.get(sub, 'id', '')
                    to_nick = enaml.get(sub, 'to-nick', '')
                    to_id = enaml.get(sub, 'to-id', '')
                    time = int(enaml.get(sub, 'time', '0'))
                    if sub.has_key('stored'):
                        stored = 1
                    else:
                        stored = 0
                    self.event('message', nick, id, to_nick, to_id, time, sub, stored)
                elif command == 'boot-user':
                    nick = enaml.get(sub, 'nick', '')
                    id = enaml.get(sub, 'id', '')
                    target_nick = enaml.get(sub, 'target-nick', '')
                    target_id = enaml.get(sub, 'target-id', '')
                    text = enaml.get(sub, 'text', '')
                    self.event('boot_user', nick, id, target_nick, target_id, text)
                elif command == 'subscriptions-from' or command == 'subscriptions':
                    if sub.has_key('approved') and type(sub['approved']) == type([]):
                        approved = sub['approved']
                    else:
                        approved = []
                    if sub.has_key('blocked') and type(sub['blocked']) == type([]):
                        blocked = sub['blocked']
                    else:
                        blocked = []
                    if sub.has_key('pending') and type(sub['pending']) == type([]):
                        pending = sub['pending']
                    else:
                        pending = []
                    if command == 'subscriptions':
                        event_name = 'subscriptions'
                    else:
                        event_name = 'subscriptions_from'
                    self.event(event_name, approved, blocked, pending)
                elif command == 'subscription-request':
                    id = enaml.get(sub, 'id', '')
                    text = enaml.get(sub, 'text', '')
                    self.event('subscription_request', id, text)
                elif command == 'no-subscription-request':
                    id = enaml.get(sub, 'id', '')
                    self.event('no_subscription_request', id)
                elif command == 'no-subscription':
                    id = enaml.get(sub, 'id', '')
                    self.event('no_subscription', id)
                elif command == 'subscription-accepted':
                    id = enaml.get(sub, 'id', '')
                    text = enaml.get(sub, 'text', '')
                    already = sub.has_key('already')
                    self.event('subscription_accepted', id, text, already)
                elif command == 'subscription-denied':
                    id = enaml.get(sub, 'id', '')
                    text = enaml.get(sub, 'text', '')
                    forever = sub.has_key('patience-exceeded')
                    self.event('subscription_denied', id, text, forever)
                elif command == 'unsubscribed-from':
                    id = enaml.get(sub, 'id', '')
                    self.event('unsubscribed_from', id);
                elif command == 'begin-response':
                    count = int(enaml.get(sub, 'count', '0'))
                    total = int(enaml.get(sub, 'total', '0'))
                    response = enaml.get(sub, 'response', '')
                    self.event('begin_response', response, count, total)
                elif command == 'end-response':
                    response = enaml.get(sub, 'response', '')
                    self.event('end_response', response)
                else:
                    self.event('unknown', command, sub)

